@@ -480,11 +480,11 @@ void pushGenericCommand(client *c, int where, int xx) {
         server.dirty++;
     }
 
-    addReplyLongLong(c, listTypeLength(lobj));
-
-    char *event = (where == LIST_HEAD) ? "lpush" : "rpush";
     signalModifiedKey(c, c->db, c->argv[1]);
+    char *event = (where == LIST_HEAD) ? "lpush" : "rpush";
     notifyKeyspaceEvent(NOTIFY_LIST, event, c->argv[1], c->db->id);
+
+    addReplyLongLong(c, listTypeLength(lobj));
 }
 
 /* LPUSH <key> <element> [<element> ...] */
@@ -608,10 +608,10 @@ void lsetCommand(client *c) {
          * already handled the growing case in listTypeTryConversionAppend()
          * above, so here we just need to try the conversion for shrinking. */
         listTypeTryConversion(o, LIST_CONV_SHRINKING, NULL, NULL);
-        addReply(c, shared.ok);
         signalModifiedKey(c, c->db, c->argv[1]);
         notifyKeyspaceEvent(NOTIFY_LIST, "lset", c->argv[1], c->db->id);
         server.dirty++;
+        addReply(c, shared.ok);
     } else {
         addReplyErrorObject(c, shared.outofrangeerr);
     }
@@ -628,13 +628,14 @@ void lsetCommand(client *c) {
  *
  * 'deleted' is an optional output argument to get an indication
  * if the key got deleted by this function. */
-void listPopRangeAndReplyWithKey(client *c, robj *o, robj *key, int where, long count, int signal, int *deleted) {
+void listPopRangeAndReplyWithKey(client *c, robj *o, robj *key, int where, long count, int *deleted) {
     long llen = listTypeLength(o);
     long rangelen = (count > llen) ? llen : count;
     long rangestart = (where == LIST_HEAD) ? 0 : -rangelen;
     long rangeend = (where == LIST_HEAD) ? rangelen - 1 : -1;
     int reverse = (where == LIST_HEAD) ? 0 : 1;
 
+    initDeferredReplyBuffer(c);
     /* We return key-name just once, and an array of elements */
     addReplyArrayLen(c, 2);
     addReplyBulk(c, key);
@@ -643,7 +644,8 @@ void listPopRangeAndReplyWithKey(client *c, robj *o, robj *key, int where, long
     /* Pop these elements. */
     listTypeDelRange(o, rangestart, rangelen);
     /* Maintain the notifications and dirty. */
-    listElementsRemoved(c, key, where, o, rangelen, signal, deleted);
+    listElementsRemoved(c, key, where, o, rangelen, deleted);
+    commitDeferredReplyBuffer(c, 1);
 }
 
 /* Extracted from `addListRangeReply()` to reply with a quicklist list.
@@ -726,12 +728,10 @@ void addListRangeReply(client *c, robj *o, long start, long end, int reverse) {
 }
 
 /* A housekeeping helper for list elements popping tasks.
- *
- * If 'signal' is 0, skip calling signalModifiedKey().
  *
  * 'deleted' is an optional output argument to get an indication
  * if the key got deleted by this function. */
-void listElementsRemoved(client *c, robj *key, int where, robj *o, long count, int signal, int *deleted) {
+void listElementsRemoved(client *c, robj *key, int where, robj *o, long count, int *deleted) {
     char *event = (where == LIST_HEAD) ? "lpop" : "rpop";
 
     notifyKeyspaceEvent(NOTIFY_LIST, event, key, c->db->id);
@@ -744,7 +744,7 @@ void listElementsRemoved(client *c, robj *key, int where, robj *o, long count, i
         listTypeTryConversion(o, LIST_CONV_SHRINKING, NULL, NULL);
         if (deleted) *deleted = 0;
     }
-    if (signal) signalModifiedKey(c, c->db, key);
+    signalModifiedKey(c, c->db, key);
     server.dirty += count;
 }
 
@@ -778,10 +778,10 @@ void popGenericCommand(client *c, int where) {
         /* Pop a single element. This is POP's original behavior that replies
          * with a bulk string. */
         value = listTypePop(o, where);
+        listElementsRemoved(c, c->argv[1], where, o, 1, NULL);
         serverAssert(value != NULL);
         addReplyBulk(c, value);
         decrRefCount(value);
-        listElementsRemoved(c, c->argv[1], where, o, 1, 1, NULL);
     } else {
         /* Pop a range of elements. An addition to the original POP command,
          *  which replies with a multi-bulk. */
@@ -791,9 +791,11 @@ void popGenericCommand(client *c, int where) {
         long rangeend = (where == LIST_HEAD) ? rangelen - 1 : -1;
         int reverse = (where == LIST_HEAD) ? 0 : 1;
 
+        initDeferredReplyBuffer(c);
         addListRangeReply(c, o, rangestart, rangeend, reverse);
         listTypeDelRange(o, rangestart, rangelen);
-        listElementsRemoved(c, c->argv[1], where, o, rangelen, 1, NULL);
+        listElementsRemoved(c, c->argv[1], where, o, rangelen, NULL);
+        commitDeferredReplyBuffer(c, 1);
     }
 }
 
@@ -823,7 +825,7 @@ void mpopGenericCommand(client *c, robj **keys, int numkeys, int where, long cou
         if (llen == 0) continue;
 
         /* Pop a range of elements in a nested arrays way. */
-        listPopRangeAndReplyWithKey(c, o, key, where, count, 1, NULL);
+        listPopRangeAndReplyWithKey(c, o, key, where, count, NULL);
 
         /* Replicate it as [LR]POP COUNT. */
         robj *count_obj = createStringObjectFromLongLong((count > llen) ? llen : count);
@@ -907,8 +909,9 @@ void ltrimCommand(client *c) {
     } else {
         listTypeTryConversion(o, LIST_CONV_SHRINKING, NULL, NULL);
     }
-    signalModifiedKey(c, c->db, c->argv[1]);
-    server.dirty += (ltrim + rtrim);
+    long long total_trim = ltrim + rtrim;
+    if (total_trim) signalModifiedKey(c, c->db, c->argv[1]);
+    server.dirty += total_trim;
     addReply(c, shared.ok);
 }
 
@@ -1116,7 +1119,7 @@ void lmoveGenericCommand(client *c, int wherefrom, int whereto) {
         value = listTypePop(sobj, wherefrom);
         serverAssert(value); /* assertion for valgrind (avoid NPD) */
         lmoveHandlePush(c, c->argv[2], dobj, value, whereto);
-        listElementsRemoved(c, touchedkey, wherefrom, sobj, 1, 1, NULL);
+        listElementsRemoved(c, touchedkey, wherefrom, sobj, 1, NULL);
 
         /* listTypePop returns an object with its refcount incremented */
         decrRefCount(value);
@@ -1190,7 +1193,7 @@ void blockingPopGenericCommand(client *c, robj **keys, int numkeys, int where, i
         if (count != -1) {
             /* BLMPOP, non empty list, like a normal [LR]POP with count option.
              * The difference here we pop a range of elements in a nested arrays way. */
-            listPopRangeAndReplyWithKey(c, o, key, where, count, 1, NULL);
+            listPopRangeAndReplyWithKey(c, o, key, where, count, NULL);
 
             /* Replicate it as [LR]POP COUNT. */
             robj *count_obj = createStringObjectFromLongLong((count > llen) ? llen : count);
@@ -1203,12 +1206,11 @@ void blockingPopGenericCommand(client *c, robj **keys, int numkeys, int where, i
         robj *value = listTypePop(o, where);
         serverAssert(value != NULL);
 
+        listElementsRemoved(c, key, where, o, 1, NULL);
         addReplyArrayLen(c, 2);
         addReplyBulk(c, key);
         addReplyBulk(c, value);
         decrRefCount(value);
-        listElementsRemoved(c, key, where, o, 1, 1, NULL);
-
         /* Replicate it as an [LR]POP instead of B[LR]POP. */
         rewriteClientCommandVector(c, 2, (where == LIST_HEAD) ? shared.lpop : shared.rpop, key);
         return;
