@@ -228,11 +228,11 @@ start_cluster 1 0 {tags {external:skip cluster} overrides {cluster-slot-stats-en
     R 0 FLUSHALL
 
     test "CLUSTER SLOT-STATS cpu-usec for blocking commands, unblocked on timeout." {
-        # Blocking command with 1 second timeout.
+        # Blocking command with 0.5 seconds timeout.
         set rd [valkey_deferring_client]
-        $rd BLPOP $key 1
+        $rd BLPOP $key 0.5
 
-        # Confirm that the client is blocked, then unblocked after 1 second timeout.
+        # Confirm that the client is blocked, then unblocked within 1 second.
         wait_for_blocked_clients_count 1
         wait_for_blocked_clients_count 0
 
@@ -474,8 +474,8 @@ start_cluster 1 1 {tags {external:skip cluster} overrides {cluster-slot-stats-en
         set slot [R 0 cluster keyslot $channel]
         set primary [Rn 0]
         set replica [Rn 1]
-        set replica_subscriber [valkey_deferring_client -1]
-        $replica_subscriber SSUBSCRIBE $channel
+        set replica_subcriber [valkey_deferring_client -1]
+        $replica_subcriber SSUBSCRIBE $channel
         # *2\r\n$10\r\nssubscribe\r\n$7\r\nchannel\r\n --> 34 bytes.
         $primary SPUBLISH $channel hello
         # *3\r\n$8\r\nspublish\r\n$7\r\nchannel\r\n$5\r\nhello\r\n --> 42 bytes.
@@ -524,9 +524,12 @@ start_cluster 1 0 {tags {external:skip cluster}} {
         R 0 SET $key value
         # +OK\r\n --> 5 bytes
 
+        R 0 GET $key
+        # $3\r\nvalue\r\n -> 11 bytes
+
         set expected_slot_stats [
             dict create $key_slot [
-                dict create network-bytes-out 5
+                dict create network-bytes-out 16
             ]
         ]
         set slot_stats [R 0 CLUSTER SLOT-STATS SLOTSRANGE 0 16383]
@@ -971,4 +974,4 @@ start_cluster 1 1 {tags {external:skip cluster} overrides {cluster-slot-stats-en
     }
     R 0 CONFIG RESETSTAT
     R 1 CONFIG RESETSTAT
-}
\ No newline at end of file
+}
